/************************************************************/
// Fort Agent binary patcher for Agent 1.9/32.560          //
// Programming by JTHZ.com, November 27, 1999, MSVC6PRO     //
// Updated to 1.8/32.553 on December 22, 2001               //
// Updated to 1.9/32.560 on February 7, 2002                //
// Updated to 1.91/32.560 on April 9, 2002                  //
// Win32 GUI app, freeware, use at own risk, blah blah blah //
/************************************************************/

#include "forte.h"

//---------------------------------------------------------------------------

static const char BACKUP_NAME[] = "Agent.xxx";

static const DWORD VERSION_OFFSET = 0x001b6b2a;
static const DWORD VERSION_NBYTES = 12;
static const BYTE VERSION_ORG[VERSION_NBYTES] = 
{
	0x25, 0x73, 0x20, 0x25, 0x73, 0x2f, 
	0x33, 0x32, 0x2e, 0x35, 0x36, 0x34
}; // %s %s/32.564

static const BYTE VERSION_MODIFIED[VERSION_NBYTES] = 
{
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20
};

static const DWORD MIME_OFFSET = 0x001f72a1;
static const DWORD MIME_NBYTES = 3;
static const BYTE MIME_ORG[MIME_NBYTES] = {0x31, 0x2e, 0x30}; // 1.0
static const BYTE MIME_MODIFIED[MIME_NBYTES] = {0x20, 0x20, 0x20};

static const DWORD SPLASH_OFFSET = 0x00257f5c;
static const DWORD SPLASH_NBYTES = 6;
static const BYTE SPLASH_ORG[SPLASH_NBYTES] = {0xf4, 0x01, 0, 0, 0x5e, 0x01};
static const BYTE SPLASH_MODIFIED[SPLASH_NBYTES] = {0x01, 0, 0, 0, 0x01, 0};

static const DWORD FOURAX_OFFSET = 0x001d66ac;
static const DWORD FOURAX_NBYTES = 8;

static const DWORD DISTRIBUTION_OFFSET = 0x001daeba;
static const DWORD DISTRIBUTION_NBYTES = 13;

static const DWORD KEYWORDS_OFFSET = 0x001daec7;
static const DWORD KEYWORDS_NBYTES = 9;

static const int nFields = 6;

static const DWORD dwOffsets[nFields] = 
{
	VERSION_OFFSET,
	MIME_OFFSET,
	SPLASH_OFFSET,
	FOURAX_OFFSET,
	DISTRIBUTION_OFFSET,
	KEYWORDS_OFFSET
};

static const DWORD dwLengths[nFields] = 
{
	VERSION_NBYTES,
	MIME_NBYTES,
	SPLASH_NBYTES,
	FOURAX_NBYTES,
	DISTRIBUTION_NBYTES,
	KEYWORDS_NBYTES
};

static const DWORD AGENT_EXE_SIZE                           = 2868768;

static const DWORD LOWLEVEL_VERSION_OFFSET                  = 0x002bc484;
static const DWORD LOWLEVEL_VERSION_NBYTES                  = 7;
static const BYTE LOWLEVEL_VERSION[LOWLEVEL_VERSION_NBYTES] = {0x31, 0x00, 0x2e, 0x00, 0x39, 0x00, 0x31};

typedef struct _AGENTDATA
{
	BYTE VersionStamp[VERSION_NBYTES];
	BYTE MIME[MIME_NBYTES];
	BYTE Splash[SPLASH_NBYTES];
	char psz4AX[FOURAX_NBYTES];
	char pszDistribution[DISTRIBUTION_NBYTES];
	char pszKeywords[KEYWORDS_NBYTES];
} AGENTDATA;

//---------------------------------------------------------------------------

// Global var. that holds fullname of valid agent.exe or "".
char szAgentExe[MAX_PATH] = ""; 

// Global var. that holds fullname of valid agent.ini or "".
char szAgentIni[MAX_PATH] = "";

//---------------------------------------------------------------------------

int WINAPI WinMain(HINSTANCE hInstance,	HINSTANCE, PSTR, int)
{
	// Only one instance.
	HANDLE hMutex = CreateMutex
	(
		0,
		TRUE,
		"{C1CF1380-A1E6-11D3-8E0E-DFAA46B85932}"
	);

	if(GetLastError() == ERROR_ALREADY_EXISTS)
	{
		if(hMutex)
		{
			CloseHandle(hMutex);
		}
		return 0;
	}

	// Run dialog.
	DialogBox
	(
		hInstance,
		MAKEINTRESOURCE(IDD_INTERFACE),
		0,
		reinterpret_cast <DLGPROC> (DialogProc)
	);

	// Cleanup.
	ReleaseMutex(hMutex);
	CloseHandle(hMutex);

	return 0;
}

//---------------------------------------------------------------------------

BOOL CALLBACK DialogProc
(
	HWND hwndDlg, 
	UINT uMsg, 
	WPARAM wParam, 
	LPARAM lParam
)
{
	switch(uMsg)
	{
		case WM_COMMAND:
			switch(LOWORD(wParam))
			{
				case IDC_APPLY:
					if(GetAsyncKeyState(VK_SHIFT) & 0x8000) // Shift down.
					{ 
						Commercial(hwndDlg);

						return TRUE;
					}
					ApplyPatch(hwndDlg, szAgentExe, szAgentIni);
					UpdateDisplayExe(hwndDlg, szAgentExe);
					UpdateDisplayIni(hwndDlg, szAgentIni);

					return TRUE;

				case IDC_FINDAGENT:
					if(FindExe(hwndDlg, szAgentExe))
					{
						UpdateDisplayExe(hwndDlg, szAgentExe);
						if(AutoFindIni(szAgentExe, szAgentIni))
						{
							UpdateDisplayIni(hwndDlg, szAgentIni);
						}
						else
						{
							lstrcpy(szAgentIni, "");
							UpdateDisplayIni(hwndDlg, 0);
						}
						UpdateDisplayApply(hwndDlg, szAgentExe, szAgentIni);
					}

					return TRUE;

				case IDC_FINDINI:
					if(FindIni(hwndDlg, szAgentExe, szAgentIni))
					{
						UpdateDisplayIni(hwndDlg, szAgentIni);
						UpdateDisplayApply(hwndDlg, szAgentExe, szAgentIni);
					}

					return TRUE;

				case IDCANCEL:
					EndDialog(hwndDlg, 0);

					return TRUE;
			}

			return FALSE;

		case WM_INITDIALOG:
			SetControlLimits(hwndDlg);
			if(AutoFindExe(hwndDlg, szAgentExe))
			{
				UpdateDisplayExe(hwndDlg, szAgentExe);

				if(AutoFindIni(szAgentExe, szAgentIni))
				{
					UpdateDisplayIni(hwndDlg, szAgentIni);
				}
				else
				{
					lstrcpy(szAgentIni, "");
					UpdateDisplayIni(hwndDlg, 0);
				}

				UpdateDisplayApply(hwndDlg, szAgentExe, szAgentIni);
			}

			return TRUE;

	}

	return FALSE;
}

//---------------------------------------------------------------------------

void SetControlLimits(HWND hwndDlg)
{
	// Set max. values for edit controls.
	// hwndDlg: dialog
	//
	const int nControls = 3;

	int nControlIDs[nControls] = 
	{
		IDC_4AX,
		IDC_DISTRIBUTION,
		IDC_KEYWORDS
	};

	DWORD dwLimits[nControls] = 
	{
		FOURAX_NBYTES,
		DISTRIBUTION_NBYTES,
		KEYWORDS_NBYTES
	};

	for(int i = 0; i < nControls; ++i)
	{
		SendDlgItemMessage
		(
			hwndDlg,
			nControlIDs[i],
			EM_SETLIMITTEXT,
			static_cast <WPARAM> (dwLimits[i] - 1L),
			0L
		);
	}
}

//---------------------------------------------------------------------------

BOOL AutoFindExe(HWND hwndDlg, PSTR pstrExe)
{
	// Attempts to find Agent in registry
	// pstrExe: global varibale to copy exe to if verified
	//
	// read value from registry
	HKEY hKey;

	LONG lResult = RegOpenKeyEx
	(
		HKEY_CURRENT_USER,
		"Software\\Forte\\Agent\\Paths",
		0,
		KEY_READ,
		&hKey
	);

	if(lResult != ERROR_SUCCESS)
	{
		return FALSE;
	}

	char szBuf[MAX_PATH];

	DWORD dwBufSize = sizeof(szBuf);

	lResult = RegQueryValueEx
	(
		hKey,
		"ExeFile",
		0,
		0,
		reinterpret_cast <PBYTE> (szBuf),
		&dwBufSize
	);

	RegCloseKey(hKey);

	// verify version
	if(VerifyExeVersion(hwndDlg, szBuf))
	{
		lstrcpy(pstrExe, szBuf); // update global variable

		return TRUE;
	}

	return FALSE;
}

//---------------------------------------------------------------------------

BOOL AutoFindIni(PCSTR pcstrExe, PSTR pstrIni)
{
	// attempts to find agent.ini, based on location of agent.exe
	// pcstrExe: fullname of exe
	// pstrIni: global varibale to copy ini to if verified

	if(!lstrlen(pcstrExe))
	{
		return FALSE;
	}

	char szBuf[MAX_PATH];

	if(ExtractFilePath(pcstrExe, szBuf))
	{
		int len = lstrlen(szBuf);

		if(len && szBuf[len - 1] != '\\')
		{
			lstrcat(szBuf, "\\");
		}

		lstrcat(szBuf, "Data\\agent.ini");

		// verify existence
		if(FileExists(szBuf))
		{
			lstrcpy(pstrIni, szBuf); // update global variable

			return TRUE;
		}
	}

	return FALSE;

}

//---------------------------------------------------------------------------

BOOL FindExe(HWND hwndDlg, PSTR pstrExe)
{
	LPMALLOC lpMalloc;

	if(SHGetMalloc(&lpMalloc) != NOERROR)
	{
		return FALSE;
	}

	char szDummy[MAX_PATH] = "";

	BROWSEINFO browseInfo;

	browseInfo.hwndOwner      = hwndDlg;
	browseInfo.pidlRoot       = 0;
	browseInfo.pszDisplayName = szDummy;
	browseInfo.lpszTitle      = "Find directory where AGENT.EXE lives";
	browseInfo.ulFlags        = BIF_RETURNFSANCESTORS |
		      			        BIF_RETURNONLYFSDIRS  |
								BIF_DONTGOBELOWDOMAIN;

	char szDir[MAX_PATH];

	if(ExtractFilePath(pstrExe, szDir))
	{ 
		// Start path is possible
		browseInfo.lpfn = reinterpret_cast <BFFCALLBACK> (BrowseStartDirectory);
		browseInfo.lParam = reinterpret_cast <LPARAM> (szDir);
	}
	else
	{
		browseInfo.lpfn   = 0;
		browseInfo.lParam = 0;
	}

	LPITEMIDLIST lpItemIDList;

	char szBuf[MAX_PATH] = "";

	if(lpItemIDList = SHBrowseForFolder(&browseInfo))
	{
		SHGetPathFromIDList(lpItemIDList, szBuf);
	}

	lpMalloc->Free(lpItemIDList);
	lpMalloc->Release();

	int len = lstrlen(szBuf);

	if(!len)
	{
		return FALSE;
	}

	if(szBuf[len - 1] != '\\')
	{
		lstrcat(szBuf, "\\");
	}

	lstrcat(szBuf, "Agent.exe");

	if(!FileExists(szBuf))
	{
		MessageBox
		(
			hwndDlg,
			"No AGENT.EXE found in this directory!",
			"ERROR",
			MB_ICONEXCLAMATION
		);

		return FALSE;
	}

	// verify version

	if(VerifyExeVersion(hwndDlg, szBuf))
	{
		lstrcpy(pstrExe, szBuf); // update global variable

		return TRUE;
	}

	return FALSE;
}

//---------------------------------------------------------------------------

BOOL FindIni(HWND hwndDlg, PCSTR pcstrExe, PSTR pstrIni)
{
	LPMALLOC lpMalloc;

	if(SHGetMalloc(&lpMalloc) != NOERROR)
	{
		return FALSE;
	}

	char szDummy[MAX_PATH] = "";

	BROWSEINFO browseInfo;

	browseInfo.hwndOwner      = hwndDlg;
	browseInfo.pidlRoot       = 0;
	browseInfo.pszDisplayName = szDummy;
	browseInfo.lpszTitle      = "Find directory where AGENT.INI lives";
	browseInfo.ulFlags        = BIF_RETURNFSANCESTORS |
			                    BIF_RETURNONLYFSDIRS  | 
								BIF_DONTGOBELOWDOMAIN;

	char szDir[MAX_PATH];

	if(ExtractFilePath(pstrIni, szDir))
	{ 
		// start path is possible
		browseInfo.lpfn = reinterpret_cast <BFFCALLBACK> (BrowseStartDirectory);
		browseInfo.lParam = reinterpret_cast <LPARAM> (szDir);
	}
	else
	{
		if(ExtractFilePath(pcstrExe, szDir))
		{ 
			// try exe-path
			browseInfo.lpfn = reinterpret_cast <BFFCALLBACK> (BrowseStartDirectory);
			browseInfo.lParam = reinterpret_cast <LPARAM> (szDir);
		}
		else
		{
			browseInfo.lpfn   = 0;
			browseInfo.lParam = 0;
		}
	}

	LPITEMIDLIST lpItemIDList;

	char szBuf[MAX_PATH] = "";

	if(lpItemIDList = SHBrowseForFolder(&browseInfo))
	{
		SHGetPathFromIDList(lpItemIDList, szBuf);
	}

	lpMalloc->Free(lpItemIDList);
	lpMalloc->Release();

	int len = lstrlen(szBuf);

	if(!len)
	{
		return FALSE;
	}

	if(szBuf[len - 1] != '\\')
	{
		lstrcat(szBuf, "\\");
	}

	lstrcat(szBuf, "Agent.ini");

	if(!FileExists(szBuf))
	{
		MessageBox
		(
			hwndDlg,
			"No AGENT.INI found in this directory!",
			"ERROR",
			MB_ICONEXCLAMATION
		);

		return FALSE;
	}

	lstrcpy(pstrIni, szBuf); // update global variable

	return TRUE;
}

//---------------------------------------------------------------------------

BOOL VerifyExeVersion(HWND hwndDlg, PCSTR pcstrExe)
{
	// verify version of Agent.exe
	// returns TRUE if pcstrExe is correct version
	// pcstrExe: name of exe to check

	if(!lstrlen(pcstrExe))
	{
		return FALSE;
	}

	HANDLE hFile = CreateFile
	(
		pcstrExe,
		GENERIC_READ,
		FILE_SHARE_READ,
		0,
		OPEN_EXISTING,
		FILE_ATTRIBUTE_NORMAL | FILE_FLAG_RANDOM_ACCESS,
		0
	);

	if(hFile == INVALID_HANDLE_VALUE)
	{
		return FALSE;
	}

	DWORD dwFileSize = GetFileSize(hFile, 0);

	SetFilePointer
	(
		hFile,
		LOWLEVEL_VERSION_OFFSET,
		0,
		FILE_BEGIN
	);

	BYTE buf[8];

	ZeroMemory(buf, sizeof(buf));

	DWORD dwBytesRead;

	ReadFile
	(
		hFile,
		reinterpret_cast <PVOID> (buf),
		LOWLEVEL_VERSION_NBYTES,
		&dwBytesRead,
		0
	);

	CloseHandle(hFile);

	if((dwFileSize == AGENT_EXE_SIZE) && 
	   (!memcmp(buf, LOWLEVEL_VERSION, LOWLEVEL_VERSION_NBYTES)))
	{
		return TRUE;
	}

	MessageBox
	(
		hwndDlg,
		"Wrong version of AGENT.EXE",
		"ERROR",
		MB_ICONEXCLAMATION
	);

	return FALSE;
}

//---------------------------------------------------------------------------

void UpdateDisplayExe(HWND hwndDlg, PCSTR pcstrExe)
{
	// update interface-stuff related to executable
	// hwndDlg: dialog to update
	// pcstrExe: name of exe to read from

	if(!FileExists(pcstrExe))
	{
		return;
	}

	// try to read current values from exe
	AGENTDATA AgentData;

	HANDLE hFile = CreateFile
	(
		pcstrExe,
		GENERIC_READ,
		FILE_SHARE_READ,
		0,
		OPEN_EXISTING,
		FILE_ATTRIBUTE_NORMAL | FILE_FLAG_RANDOM_ACCESS,
		0
	);

	if(hFile == INVALID_HANDLE_VALUE)
	{
		MessageBox
		(
			hwndDlg,
			"Error opening AGENT.EXE",
			"ERROR",
			MB_ICONEXCLAMATION
		);

		return;
	}

	ZeroMemory(&AgentData, sizeof(AGENTDATA));

	PVOID pData[nFields] = 
	{
		reinterpret_cast <PVOID> (AgentData.VersionStamp),
		reinterpret_cast <PVOID> (AgentData.MIME),
		reinterpret_cast <PVOID> (AgentData.Splash),
		reinterpret_cast <PVOID> (AgentData.psz4AX),
		reinterpret_cast <PVOID> (AgentData.pszDistribution),
		reinterpret_cast <PVOID> (AgentData.pszKeywords)
	};

	DWORD dwBytesRead;

	for(int i = 0 ; i < nFields; ++i)
	{
		SetFilePointer
		(
			hFile,
			dwOffsets[i],
			0,
			FILE_BEGIN
		);

		ReadFile
		(
			hFile,
			pData[i],
			dwLengths[i],
			&dwBytesRead,
			0
		);
	}

	CloseHandle(hFile);

	// button's caption becomes name of exe 
	SetWindowText
	(
		GetDlgItem(hwndDlg, IDC_FINDAGENT),
		pcstrExe
	);

	// enable controls
	const int nControls = 9;

	int nControlIDs[nControls] = 
	{
        IDC_SPLASH,
        IDC_PRODUCT,
        IDC_MIME,
		IDC_STATIC1,
		IDC_STATIC2,
        IDC_DISTRIBUTION,
        IDC_KEYWORDS,
        IDC_4AX,
		IDC_STATIC3
	};

	for(i = 0; i < nControls; ++i)
	{
		EnableWindow
		(
			GetDlgItem(hwndDlg, nControlIDs[i]),
			TRUE
		);
	}

	// set edit-control contents to current data
	SetDlgItemText(hwndDlg, IDC_4AX, AgentData.psz4AX);
	SetDlgItemText(hwndDlg, IDC_DISTRIBUTION, AgentData.pszDistribution);
	SetDlgItemText(hwndDlg, IDC_KEYWORDS, AgentData.pszKeywords);

	// is original version still in exe?
	CheckDlgButton
	(
		hwndDlg, 
		IDC_PRODUCT,
		memcmp(VERSION_ORG, AgentData.VersionStamp, VERSION_NBYTES)
	);

	// is splashscreen still enabled?
	CheckDlgButton
	(
		hwndDlg, 
		IDC_SPLASH,
		memcmp(SPLASH_ORG, AgentData.Splash, SPLASH_NBYTES)
	);

	// is MIME version still in exe?
	CheckDlgButton
	(
		hwndDlg, 
		IDC_MIME,
		memcmp(MIME_ORG, AgentData.MIME, MIME_NBYTES)
	);

	// backup already exists?
	char szBuf[MAX_PATH];

	if(ExtractFilePath(pcstrExe, szBuf))
	{
		int len = lstrlen(szBuf);
		if(len && szBuf[len - 1] != '\\')
		{
			lstrcat(szBuf, "\\");
		}

		lstrcat(szBuf, BACKUP_NAME);

		BOOL bBackupExists = FileExists(szBuf);

		EnableWindow
		(
			GetDlgItem(hwndDlg, IDC_BACKUP),
			!bBackupExists
		);

		CheckDlgButton(
			hwndDlg, 
			IDC_BACKUP,
			!bBackupExists
		);
	}

}

//---------------------------------------------------------------------------

void UpdateDisplayIni(HWND hwndDlg, PCSTR pcstrIni)
{
	// update interface-stuff related to ini
	// hwndDlg: dialog to update
	// pcstrIni: name of ini to read from or 0 (disable ini-interface stuff)

	BOOL bEnable = lstrlen(pcstrIni);

	SetWindowText
	(
		GetDlgItem(hwndDlg, IDC_FINDINI),
		bEnable ? pcstrIni : "Click to find Agent.ini..."
	);

	EnableWindow
	(
		GetDlgItem(hwndDlg, IDC_GENID),
		bEnable
	);

	EnableWindow
	(
		GetDlgItem(hwndDlg, IDC_ALLFIELDS),
		bEnable
	);

	EnableWindow
	(
		GetDlgItem(hwndDlg, IDC_STATIC4),
		bEnable
	);

	CheckDlgButton
	(
		hwndDlg, 
		IDC_GENID,
		bEnable ? GetIni(pcstrIni, "Message", "GenerateId", 0) == 1 : FALSE
	);

	CheckDlgButton
	(
		hwndDlg, 
		IDC_ALLFIELDS,
		bEnable ? GetIni(pcstrIni, "Message", "ShowAllFields", 0) : FALSE
	);
}

//---------------------------------------------------------------------------

void UpdateDisplayApply(HWND hwndDlg, PCSTR pcstrExe, PCSTR pcstrIni)
{
	// turn apply button on or off
	// hwndDlg: dialog to update
	// pcstrExe: executable - 1 of these must exist if Apply is enabled
	// pcstrIni: ini        - 1 of these must exist if Apply is enabled

	EnableWindow
	(
		GetDlgItem(hwndDlg, IDC_APPLY),
		FileExists(pcstrExe) || FileExists(pcstrIni)
	);
}

//---------------------------------------------------------------------------

void Commercial(HWND hwndDlg)
{
	MessageBox(
		hwndDlg,
		"Created by\r\nJ.B.Thyssen\r\nH.Zimmerman",
		"https://JTHZ.com",
		MB_ICONINFORMATION
	);
}

//---------------------------------------------------------------------------

BOOL FileExists(PCSTR pcstrFileName)
{
	if(!lstrlen(pcstrFileName))
	{
		return FALSE;
	}

	HANDLE hFile = CreateFile
	(
		pcstrFileName,
		GENERIC_READ,
		FILE_SHARE_READ,
		0,
		OPEN_EXISTING,
		FILE_ATTRIBUTE_NORMAL,
		0
	);

	if(hFile == INVALID_HANDLE_VALUE)
	{
		return FALSE;
	}

	CloseHandle(hFile);

	return TRUE;
}

//---------------------------------------------------------------------------

BOOL ExtractFilePath(PCSTR pcstrExe, PSTR pstrBuffer)
{
	lstrcpy(pstrBuffer, pcstrExe);

	if(char *p = strrchr(pstrBuffer, '\\'))
	{
		*p = 0;

		int len = lstrlen(pstrBuffer);

		if(len && pstrBuffer[len - 1] == ':')
		{
			lstrcat(pstrBuffer, "\\");
		}

		return TRUE;
	}

	return FALSE;
}

//---------------------------------------------------------------------------

/* Old version:

BOOL DirectoryExists(PCSTR pcstrDir)
{
	int len = lstrlen(pcstrDir);

	if(!len)
	{
		return FALSE;
	}

	char szOldDir[MAX_PATH];

	GetCurrentDirectory(sizeof(szOldDir), szOldDir);

	BOOL bResult = SetCurrentDirectory(pcstrDir);

	SetCurrentDirectory(szOldDir);

	return bResult;
}
*/

// Rewrite 02072002, better, stronger, faster!

BOOL DirectoryExists(PCSTR pcstrDir)
{

  DWORD Code = GetFileAttributes(pcstrDir);

  // Returnvalue in case of an error.
  static const DWORD ErrorReturn = 0xFFFFFFFF;

  return (Code != ErrorReturn) && (Code & FILE_ATTRIBUTE_DIRECTORY);

}


//---------------------------------------------------------------------------

int CALLBACK BrowseStartDirectory
(
	HWND hwnd,
	UINT uMsg,
	LPARAM wParam,
	LPARAM lParam
)
{
	if(uMsg == BFFM_INITIALIZED)
	{
		PCSTR pcstrDir = reinterpret_cast <PCSTR> (lParam);

		if(DirectoryExists(pcstrDir))
		{
			SendMessage
			(
				hwnd,
				BFFM_SETSELECTION,
				static_cast <LPARAM> (1),
				lParam
			);
		}
	}

	return 0;
}

//---------------------------------------------------------------------------

int GetIni
(
	PCSTR pcstrIni, 
	PCSTR pcstrSection, 
	PCSTR pcstrKey, 
	int nDefaultValue
)
{
	if(lstrlen(pcstrIni) && FileExists(pcstrIni))
	{
		return GetPrivateProfileInt
		(
			pcstrSection,
			pcstrKey,
			nDefaultValue,
			pcstrIni
		);
	}

	return nDefaultValue;
}

//---------------------------------------------------------------------------

void SetIni
(
	PCSTR pcstrIni, 
	PCSTR pcstrSection, 
	PCSTR pcstrKey, 
	int nValue
)
{
	if(lstrlen(pcstrIni) && FileExists(pcstrIni))
	{
		char buf[16];

		wsprintf(buf, "%d", nValue);

		WritePrivateProfileString
		(
			pcstrSection,
			pcstrKey,
			buf,
			pcstrIni
		);
	}
}

//---------------------------------------------------------------------------

void ApplyPatch(HWND hwndDlg, PCSTR pcstrExe, PCSTR pcstrIni)
{
	HCURSOR hOldCursor = SetCursor(LoadCursor(0, IDC_WAIT));

	// ini stuff

	if(FileExists(pcstrIni))
	{
		const char PATCH_SECTION[] = "patch undo";

		if(IsDlgButtonChecked(hwndDlg, IDC_GENID))
		{
			int nOld = GetIni(pcstrIni, "Message", "GenerateId", 0);

			if(nOld != 1)
			{
				SetIni(pcstrIni, PATCH_SECTION, "GenerateId", nOld);
			}

			SetIni(pcstrIni, "Message", "GenerateId", 1);
		}
		else
		{
			int nOld = GetIni(pcstrIni, PATCH_SECTION, "GenerateId", 0);

			SetIni(pcstrIni, "Message", "GenerateId", nOld);
		}

		SetIni
		(
			pcstrIni, 
			"Message", 
			"ShowAllFields", 
			(IsDlgButtonChecked(hwndDlg, IDC_ALLFIELDS) ? 1 : 0)
		);
	}

	if(!FileExists(pcstrExe))
	{
		return; // skip rest, ini-patch only
	}

	// try to read current values from exe
	HANDLE hFile = CreateFile
	(
		pcstrExe,
		GENERIC_READ,
		FILE_SHARE_READ,
		0,
		OPEN_EXISTING,
		FILE_ATTRIBUTE_NORMAL | FILE_FLAG_RANDOM_ACCESS,
		0
	);

	if(hFile == INVALID_HANDLE_VALUE)
	{
		MessageBox
		(
			hwndDlg,
			"Error opening AGENT.EXE",
			"ERROR",
			MB_ICONEXCLAMATION
		);
		return;
	}

	AGENTDATA adOriginal;

	ZeroMemory(&adOriginal, sizeof(AGENTDATA));

	PVOID pOriginalData[nFields] = 
	{
		reinterpret_cast <PVOID> (adOriginal.VersionStamp),
		reinterpret_cast <PVOID> (adOriginal.MIME),
		reinterpret_cast <PVOID> (adOriginal.Splash),
		reinterpret_cast <PVOID> (adOriginal.psz4AX),
		reinterpret_cast <PVOID> (adOriginal.pszDistribution),
		reinterpret_cast <PVOID> (adOriginal.pszKeywords)
	};

	DWORD dwBytesRead;

	for(int i = 0 ; i < nFields; ++i)
	{
		SetFilePointer
		(
			hFile,
			dwOffsets[i],
			0,
			FILE_BEGIN
		);

		ReadFile
		(
			hFile,
			pOriginalData[i],
			dwLengths[i],
			&dwBytesRead,
			0
		);
	}

	CloseHandle(hFile);

	// create second struct to compare and fill with data from dialog

	AGENTDATA adProposed;

	ZeroMemory(&adProposed, sizeof(AGENTDATA));

	CopyMemory
	(
		adProposed.VersionStamp,
		IsDlgButtonChecked
		(
			hwndDlg,
			IDC_PRODUCT
		) ? VERSION_MODIFIED : VERSION_ORG,
		VERSION_NBYTES
	);

	CopyMemory
	(
		adProposed.Splash,
		IsDlgButtonChecked
		(
			hwndDlg,
			IDC_SPLASH
		) ? SPLASH_MODIFIED : SPLASH_ORG,
		SPLASH_NBYTES
	);

	CopyMemory
	(
		adProposed.MIME,
		IsDlgButtonChecked
		(
			hwndDlg,
			IDC_MIME
		) ? MIME_MODIFIED : MIME_ORG,
		MIME_NBYTES
	);

	GetDlgItemText
	(
		hwndDlg,
		IDC_4AX,
		adProposed.psz4AX,
		FOURAX_NBYTES
	);

	GetDlgItemText
	(
		hwndDlg,
		IDC_DISTRIBUTION,
		adProposed.pszDistribution,
		DISTRIBUTION_NBYTES
	);

	GetDlgItemText
	(
		hwndDlg,
		IDC_KEYWORDS,
		adProposed.pszKeywords,
		KEYWORDS_NBYTES
	);

	PVOID pProposedData[nFields] = 
	{
		reinterpret_cast <PVOID> (adProposed.VersionStamp),
		reinterpret_cast <PVOID> (adProposed.MIME),
		reinterpret_cast <PVOID> (adProposed.Splash),
		reinterpret_cast <PVOID> (adProposed.psz4AX),
		reinterpret_cast <PVOID> (adProposed.pszDistribution),
		reinterpret_cast <PVOID> (adProposed.pszKeywords)
	};

	// has anything changed at all?
	BOOL bDirty = FALSE;

	for(i = 0; i < nFields; ++i)
	{
		if(memcmp(pProposedData[i], pOriginalData[i], dwLengths[i]))
		{
			bDirty = TRUE;

			break;
		}
	}

	if(!bDirty)
	{
		SetCursor(hOldCursor);

		return;
	}

	// save backup of exe first (if none exists)
	if(IsDlgButtonChecked(hwndDlg, IDC_BACKUP))
	{
		char szBuf[MAX_PATH];

		if(ExtractFilePath(pcstrExe, szBuf))
		{
			int len = lstrlen(szBuf);

			if(len && szBuf[len - 1] != '\\')
			{
				lstrcat(szBuf, "\\");
			}

			lstrcat(szBuf, BACKUP_NAME);

			CopyFile(pcstrExe, szBuf, TRUE);
		}
	}

	// change exe
	hFile = CreateFile
	(
		pcstrExe,
		GENERIC_WRITE,
		0, // deny sharing
		0,
		OPEN_EXISTING,
		FILE_ATTRIBUTE_NORMAL | FILE_FLAG_RANDOM_ACCESS,
		0
	);

	if(hFile == INVALID_HANDLE_VALUE)
	{
		SetCursor(hOldCursor);

		MessageBox
		(
			hwndDlg,
			"Error opening AGENT.EXE (is it running?)",
			"ERROR (no write access)",
			MB_ICONEXCLAMATION
		);

		return;
	}

	DWORD dwBytesWritten;

	for(i = 0; i < nFields; ++i)
	{
		if(memcmp(pProposedData[i], pOriginalData[i], dwLengths[i]))
		{
			SetFilePointer
			(
				hFile,
				dwOffsets[i],
				0,
				FILE_BEGIN
			);

			WriteFile
			(
				hFile,
				pProposedData[i],
				dwLengths[i],
				&dwBytesWritten,
				0
			);
		}
	}

	CloseHandle(hFile);

	SetCursor(hOldCursor);
}

//---------------------------------------------------------------------------
